
from __future__ import division, print_function

from dipy.workflows.denoise import NLMeansFlow
from dipy.workflows.flow_runner import run_flow

if __name__ == "__main__":
    run_flow(NLMeansFlow())
